import re
from dataclasses import dataclass
from string import Template
from typing import List

import toml

LAMPIP_CONFIG_TOML_TEMPLATE = Template(
    """\
[lampip]
layername = "${layername}"
description = ""
pyversions = ["3.7", "3.8", "3.9"]

[lampip.shrink]
compile = true
compile_optimize_level = 2
remove_dist_info = true

[lampip.shrink.plotly]
remove_jupyterlab_plotly = true
remove_data_docs = true
"""
)


def validate_layername(layername):
    if not re.search(r"^[a-zA-Z0-9_-]+$", layername):
        raise ValueError(
            f"Invalid layername `{layername}`;"
            " The layer name can contain only letters, numbers, hyphens, and underscores."
        )
    if len(layername) > 64 - 5:
        raise ValueError(
            f"Invalid layername `{layername}';" " The maximum length is 59 characters."
        )


def validate_pyversions(pyversions):
    if len(pyversions) == 0:
        raise ValueError("Invalid pyversions; pyversions should not be empty.")
    if not set(pyversions) <= {"3.9", "3.8", "3.7"}:
        raise ValueError(
            f'Invalid pyversions {pyversions}; Supported pyversions are "3.7", "3.8", "3.9" .'
        )


@dataclass
class ShrinkPlotly:
    remove_jupyterlab_plotly: bool
    remove_data_docs: bool

    @classmethod
    def from_dict(cls, dct):
        return cls(
            remove_jupyterlab_plotly=dct.get("remove_jupyterlab_plotly", True),
            remove_data_docs=dct.get("remove_data_docs", True),
        )


@dataclass
class Shrink:
    compile: bool
    compile_optimize_level: int
    remove_dist_info: bool
    plotly: ShrinkPlotly

    @classmethod
    def from_dict(cls, dct):
        return cls(
            compile=dct.get("compile", True),
            compile_optimize_level=dct.get("compile_optimize_level", 2),
            remove_dist_info=dct.get("remove_dist_info", True),
            plotly=ShrinkPlotly.from_dict(dct.get("plotly", dict())),
        )


@dataclass
class Config:
    layername: str
    pyversions: List[str]
    description: str
    shrink: Shrink

    def __post_init__(self):
        validate_layername(self.layername)
        validate_pyversions(self.pyversions)

    @classmethod
    def from_dict(cls, dct):
        return cls(
            layername=dct["layername"],
            pyversions=dct["pyversions"],
            description=dct["description"],
            shrink=Shrink.from_dict(dct.get("shrink", dict())),
        )

    @classmethod
    def load_toml(cls, toml_file: str):
        with open(toml_file, "rt") as fp:
            contents = toml.load(fp)
        kargs = contents["lampip"]
        return cls.from_dict(kargs)



from typing import Dict

import boto3
from botocore.exceptions import ClientError
from termcolor import cprint

STACKNAME = "lampip"

TEMPLATE = """\
AWSTemplateFormatVersion: 2010-09-09
Resources:
  DeploymentBucket:
    Type: 'AWS::S3::Bucket'
"""


def _create_stack_if_not_exists():
    kargs = {"TemplateBody": TEMPLATE, "Capabilities": ["CAPABILITY_IAM"]}
    cf = boto3.resource("cloudformation")
    stack = cf.Stack(STACKNAME)
    try:
        _ = stack.stack_id
        # if exists
    except ClientError:
        # if not exists
        cprint(f"Create the new cloudformation stack: {STACKNAME}", color="green")
        cf.create_stack(StackName=STACKNAME, **kargs)
        print("waiting stack create completed ...")
        cfcli = boto3.client("cloudformation")
        waiter = cfcli.get_waiter("stack_create_complete")
        waiter.wait(StackName=STACKNAME)
        print("DONE")


def get_cf_resources() -> Dict:
    _create_stack_if_not_exists()
    cf = boto3.resource("cloudformation")
    return {
        "DeploymentBucketName": cf.StackResource(
            STACKNAME, "DeploymentBucket"
        ).physical_resource_id
    }



import os
import os.path as op
import shutil
import sys
from datetime import datetime
from tempfile import TemporaryDirectory

import boto3
import sh
from termcolor import cprint

from .cloudformation import get_cf_resources
from .config import Config


def _docker(*args):
    cprint("$", "red", end=" ")
    cprint("docker " + " ".join(args), "green")
    p = None
    try:
        p = sh.docker(*args, _out=sys.stdout, _err=sys.stderr, _bg=True)
        p.wait()
    except KeyboardInterrupt:
        if p is not None:
            p.signal(2)
        sys.exit(2)


def _target_zip_basename(layername: str, ver: str) -> str:
    ts = round(datetime.utcnow().timestamp())
    return f"{layername}_{ts}_{ver}.zip"


def _build_cmd(zipbasename, cfg: Config):
    chains = []

    # ==========================================
    # COPY files on /volumepoint/other_resources
    # ==========================================
    chains.extend(["(test -d /other_resources && cp -RT /other_resources . || echo)"])

    # ===========
    # PIP INSTALL
    # ===========
    chains.extend(
        [
            "mkdir -p python",
            "pip3 install --no-cache-dir -r /volumepoint/requirements.txt -t python",
            "cd python",
        ]
    )

    # ====================================
    # REMOVE UNNECESSARY SNIPPETS OR FILES
    # ====================================
    if cfg.shrink.plotly.remove_jupyterlab_plotly:
        chains.append("(test -d jupyterlab_plotly && rm -rf jupyterlab_plotly || echo)")
    if cfg.shrink.plotly.remove_data_docs:
        chains.append(
            r'''(test -d plotly && (find plotly/validators -name '*.py' | xargs sed --in-place -z -E -e 's/kwargs\.pop\([[:space:]]*"data_docs",[[:space:]]*""".*""",?[[:space:]]*\)/kwargs.pop("data_docs", "")/') || echo)'''  # noqa
        )
    if cfg.shrink.remove_dist_info:
        chains.append("(find . -name '*.dist-info' | xargs rm -rf)")
    chains.append("(find . -name '__pycache__' | xargs rm -rf)")

    # =====
    # PATCH
    # =====
    if cfg.shrink.compile and cfg.shrink.compile_optimize_level >= 2:
        # Compile with optimize_level=2 remove docstrings on python code; and __doc__ attributes gets to be None,
        # Affected by this behavior some codes raises errors.
        chains.append(
            r"""(test -d numpy && (find numpy -name '*.py' | xargs sed --in-place -e 's/dispatcher\.__doc__/""/g') || echo)"""  # noqa
        )

    # =======
    # COMPILE
    # =======
    if cfg.shrink.compile:
        chains.extend(
            [
                f"""python -c 'import compileall; compileall.compile_dir(".", maxlevels=20, optimize={cfg.shrink.compile_optimize_level}, force=True, legacy=True, quiet=2)'""",  # noqa
                "(find . -name '*.py' | xargs rm -rf)",
            ]
        )

    # ===
    # ZIP
    # ===
    chains.extend(
        [
            "cd ..",
            "mkdir -p /volumepoint/dist",
            f"zip -r9 --quiet /volumepoint/dist/{zipbasename} .",
        ]
    )

    return " && ".join(chains)


def _make_package(zipbasename: str, ver: str, cfg: Config):
    """Run `pip install' in the docker container and zip artifacts."""
    cprint(f"Start to make {op.join('dist', zipbasename)}", "green")
    if op.exists(op.join("dist", zipbasename)):
        print(f"SKIP: {op.join('dist', zipbasename)} already exists")
        return
    with TemporaryDirectory() as tmpdir:
        curdir = op.abspath(os.curdir)
        # In order to follow symbolic links, copy files on ./other_resources to tmpdir
        os.makedirs(op.join(curdir, "other_resources"), exist_ok=True)
        other_resources_copy_dir = op.join(tmpdir, "other_resources_copy_dir")
        shutil.copytree(
            op.join(curdir, "other_resources"), other_resources_copy_dir, symlinks=False
        )
        _docker(
            "run",
            "--platform",
            "linux/amd64",
            "--rm",
            "-v",
            f"{curdir}:/volumepoint",
            "-v",
            f"{other_resources_copy_dir}:/other_resources",
            f"lambci/lambda:build-python{ver}"
            if ver != "3.9"
            else f"mlupin/docker-lambda:python{ver}-build",
            "sh",
            "-c",
            _build_cmd(zipbasename, cfg),
        )
    print(f"DONE: {op.join('dist', zipbasename)} created")


def _full_layername(layername: str, ver: str):
    suffix = "-py" + ver.replace(".", "")
    return layername + suffix


def _upload_package(zipbasename: str, ver: str, full_layername: str, description: str):
    cf_resources = get_cf_resources()
    s3 = boto3.resource("s3")
    cprint(f"Start to upload {op.join('dist', zipbasename)}", "green")

    bucketname = cf_resources["DeploymentBucketName"]
    s3.Bucket(bucketname).upload_file(op.join("dist", zipbasename), zipbasename)
    print(f"Put the package file: s3://{bucketname}/{zipbasename}")

    lambdafunc = boto3.client("lambda")
    res = lambdafunc.publish_layer_version(
        LayerName=full_layername,
        Description=description,
        Content={
            "S3Bucket": bucketname,
            "S3Key": zipbasename,
        },
        CompatibleRuntimes=[f"python{ver}"],
    )
    print(f"Publish the custom layer: {res['LayerVersionArn']}")

    print(f"DONE: {op.join('dist', zipbasename)} created")


def deploy_package(cfg: Config, upload_also=True):
    """Run make_package & upload_package according in accordance with config."""
    for ver in cfg.pyversions:
        # ex. NAME_MD5SUM_py3x.zip
        zipbasename = _target_zip_basename(cfg.layername, ver)
        # ex. NAME-py3x
        fullname = _full_layername(cfg.layername, ver)
        _make_package(zipbasename, ver, cfg)
        if upload_also:
            _upload_package(zipbasename, ver, fullname, cfg.description)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lampip', 'lampip.core']

package_data = \
{'': ['*']}

install_requires = \
['boto3', 'click', 'sh', 'termcolor', 'toml']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses']}

entry_points = \
{'console_scripts': ['lampip = lampip.entrypoint:main']}

setup_kwargs = {
    'name': 'lampip',
    'version': '0.3.0',
    'description': 'Simple CLI tool for creating custom python lambda layers',
    'long_description': '# lampip: Simple CLI tool for creating python custom lambda layers\n\n[![image](https://img.shields.io/pypi/l/lampip)](https://python.org/pypi/lampip)\n[![image](https://img.shields.io/pypi/v/lampip)](https://python.org/pypi/lampip)\n[![image](https://img.shields.io/pypi/pyversions/lampip)](https://python.org/pypi/lampip)\n[![image](https://github.com/hayashiya18/lampip/actions/workflows/pytest.yml/badge.svg)](https://github.com/hayashiya18/lampip/actions/workflows/pytest.yml)\n\n---\n\n## Features\n\n- Build Python(3.7, 3.8, 3.9)-compatible custom lambda layers using Docker and pip, and push it to AWS.\n\n- Reduce the package size using some approachs.\n  - [Byte-Compile](https://docs.python.org/3.8/library/compileall.html) (that remove source comments and docstrings).\n  - Remove `*.dist-info`.\n\n## Requirements\n\n- Python3\n- Docker\n- AWS Account\n\n## Installation\n\nYou can obtain this packages using pip.\n\n```console\n$ pip3 install lampip\n```\n\nThen you can use `lampip` command.\n\n```console\n$ lampip --help\nUsage: lampip [OPTIONS] COMMAND [ARGS]...\n\n  Simple CLI tool for creating python custom lambda layers\n\nOptions:\n  --version  Show the version and exit.\n  --help     Show this message and exit.\n\nCommands:\n  deploy  Build and push lambda layer\n  new     Create the scaffold\n\n```\n\n## Usage\n\nAt the first create the scaffold\n\n```console\n$ lampip new science\n```\n\nGo to the generated directory\n\n```console\n$ cd science\n$ ls\nlampip-config.toml  other_resources  requirements.txt\n```\n\nEdit `requirements.txt`\n\n```text\nnumpy\nscipy\npandas\n```\n\nEdit `lampip-config.toml`\n\n```toml\n[lampip]\nlayername = "science"\ndescription = "numpy, scipy, and pandas"\npyversions = ["3.7", "3.8", "3.9"]\n\n[lampip.shrink]\ncompile = true\ncompile_optimize_level = 2\nremove_dist_info = true\n\n# [lampip.shrink.plotly]\n# remove_jupyterlab_plotly = true\n# remove_data_docs = true\n```\n\nBefore you deploy the lambda layer, be sure you have AWS credentials configured.\n\n```console\n(If you do not configure AWS credentials yet, ...)\n$ aws configure\nAWS Access Key ID: ?????\nAWS Secret Acess Key: ?????\nDefault region name: ?????\n```\n\n(Option) You can switch the aws credentials using environments variables.\n\n```console\n(Case1: Using AWS CLI profile)\n$ export AWS_PROFILE="subaccount"\n\n(Case2: Using AWS access key directly)\n$ export AWS_ACCESS_KEY_ID=????\n$ export AWS_SECRET_ACCESS_KEY=?????\n$ export AWS_DEFAULT_REGION=?????\n```\n\nDeploy\n\n```console\n$ lampip deploy\nStart to make dist/science_1631253196_3.7.zip\n...\nPublish the custom layer: arn:aws:lambda:ap-northeast-1:XXXXXXXXXXXX:layer:science-py38:1\nDONE: dist/science_1631253312_3.8.zip created\n\n$ ls -lh dist\n-rw-r--r-- 1 root root 73M  9月 10 23:54 science_1631253254_3.7.zip\n-rw-r--r-- 1 root root 73M  9月 10 23:55 science_1631253312_3.8.zip\n...\n\n\n(The --no-upload option suppress uploading zip files)\n$ lampip deploy --no-upload\n```\n\nThen you can check deployed layers on AWS Console.\n\n![image](./assets/lambda_console.webp)\n',
    'author': 'hayashiya18',
    'author_email': 'sei8haya@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/hayashiya18/lampip',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import click

from . import __version__
from .core.workspace import Workspace


@click.group(help="Simple CLI tool for creating python custom lambda layers")
@click.version_option(__version__)
def main():
    pass


@main.command(help="Create the scaffold")
@click.argument("layer_name")
def new(layer_name: str):
    _ = Workspace.create_scaffold(layer_name)


@main.command(help="Build and push lambda layer")
@click.option("--upload/--no-upload", default=True)
def deploy(upload):
    ws = Workspace.load_directory(".")
    ws.deploy(upload_also=upload)



import os
import os.path as op

from termcolor import cprint

from .config import LAMPIP_CONFIG_TOML_TEMPLATE, Config, validate_layername
from .package import deploy_package


class Workspace:
    def __init__(self, directory: str, layername: str):
        self.directory = directory
        self.layername = layername

    @classmethod
    def create_scaffold(cls, layername) -> "Workspace":
        """Create the scaffold to the specified directory

        Files
        - requirements.txt: empty file
        - lampip-config.toml
        - other_layer_resources/bin
        - other_layer_resources/lib
        - other_layer_resources/python

        """
        validate_layername(layername)
        directory = op.join(".", layername)
        if op.exists(directory):
            raise FileExistsError(f"{directory} already exists")
        cprint(f"Create the scaffold: {layername}", color="green")
        os.mkdir(directory)
        # requirements.txt: empty file
        with open(op.join(directory, "requirements.txt"), "wt") as fp:
            pass
        # lampip-config.toml
        with open(op.join(directory, "lampip-config.toml"), "wt") as fp:
            fp.write(LAMPIP_CONFIG_TOML_TEMPLATE.substitute(layername=layername))
        # other_resources/bin
        # other_resources/lib
        # other_resources/python
        os.makedirs(op.join(directory, "other_resources", "bin"))
        os.makedirs(op.join(directory, "other_resources", "lib"))
        os.makedirs(op.join(directory, "other_resources", "python"))

        print(
            f"+ {op.join(directory, 'requirements.txt')}\n"
            f"+ {op.join(directory, 'lampip-confing.toml')}\n"
            f"+ {op.join(directory, 'other_resources', 'bin')}\n"
            f"+ {op.join(directory, 'other_resources', 'lib')}\n"
            f"+ {op.join(directory, 'other_resources', 'python')}"
        )

        return cls(directory, layername)

    @classmethod
    def load_directory(cls, directory: str) -> "Workspace":
        os.chdir(directory)
        config = Config.load_toml(op.join(directory, "lampip-config.toml"))
        return cls(directory, config.layername)

    def deploy(self, upload_also=True):
        """Build and upload the lambda custom layer."""
        config = Config.load_toml(op.join(self.directory, "lampip-config.toml"))
        deploy_package(config, upload_also)
